// Copyright 2020 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:provider/provider.dart';
import 'package:vm_service/vm_service.dart';

import 'auto_dispose_mixin.dart';
import 'instance_tree_controller.dart';
import 'table.dart';
import 'table_data.dart';
import 'trees.dart';
import 'utils.dart';

/// Name for predefined Reference and FieldNode should never be
/// seen by user.
const emptyName = '<empty>';
const sentinelName = '<sentinel>';

class InstanceTreeView extends StatefulWidget {
  const InstanceTreeView(
      {this.instance, this.disposeListener, this.disposeCallback});

  @override
  InstanceTreeViewState createState() => InstanceTreeViewState();

  final Instance instance;
  final ValueListenable disposeListener;
  final Function disposeCallback;
}

/// Table of the fields of an instance (type, name and value).
class InstanceTreeViewState extends State<InstanceTreeView>
    with AutoDisposeMixin {
  InstanceTreeController controller = InstanceTreeController();

  final TreeColumnData<FieldNode> treeColumn = _FieldTypeColumn();
  final List<ColumnData<FieldNode>> columns = [];

  @override
  void initState() {
    super.initState();

    // Setup table column names for instance viewer.
    columns.addAll([
      treeColumn,
      _FieldNameColumn(),
      _FieldValueColumn(),
    ]);
  }

  /*@override
  void didChangeDependencies() {
    super.didChangeDependencies();

    /*final newController = Provider.of<InstanceTreeController>(context);
    if (newController == controller) return;
    controller = newController;
    */
    cancel();

    // TODO(terry): setState should be called to set our state not change the
    //              controller. Have other ValueListenables on controller to
    //              listen to, so we don't need the setState calls.
    // Update the chart when the memorySource changes.
    // TODO(bkonyi): we were calling setState here before. Does that still work?
    addAutoDisposeListener(widget.disposeListener, widget.disposeCallback);
  }*/

  @override
  Widget build(BuildContext context) {
    print('Instance Fields: ${widget.instance.fields}');
    return TreeTable<FieldNode>(
      dataRoots: [
        for (final field in widget.instance.fields)
          FieldNode.createObject(
            controller,
            field.value,
            field.decl.type,
            field.decl.name,
          ),
      ],
      columns: columns,
      treeColumn: treeColumn,
      keyFactory: (typeRef) => PageStorageKey<String>(typeRef.name),
      sortColumn: columns[0],
      sortDirection: SortDirection.ascending,
    );
  }
}

class FieldNode extends TreeNode<FieldNode> {
  FieldNode._empty()
      : controller = null,
        instance = null,
        name = emptyName,
        isScaler = false,
        isObject = false,
        type = null,
        value = null;

  FieldNode._sentinel()
      : controller = null,
        instance = null,
        name = sentinelName,
        isScaler = false,
        isObject = false,
        type = null,
        value = null;

  FieldNode.createScaler(
      this.controller, this.instance, this.type, this.name, this.value)
      : isScaler = true,
        isObject = false;

  FieldNode.createObject(
    this.controller,
    this.instance,
    this.type,
    this.name, {
    this.onExpand,
  })  : value = null,
        isScaler = false,
        isObject = true;

  static FieldNode empty = FieldNode._empty();

  bool get isEmptyReference => this == empty;

  static FieldNode sentinel = FieldNode._sentinel();

  bool get isSentinelReference => this == sentinel;

  final InstanceTreeController controller;

  final Instance instance;

  final String type;

  final String name;

  final String value;

  final bool isScaler;

  final bool isObject;

  Function onExpand;

  @override
  void expand() {
    if (onExpand != null) {
      onExpand(this);
    }
    super.expand();
  }
}

class _FieldTypeColumn extends TreeColumnData<FieldNode> {
  _FieldTypeColumn() : super('Type');

  @override
  dynamic getValue(FieldNode dataObject) =>
      dataObject.isEmptyReference || dataObject.isSentinelReference
          ? ''
          : dataObject.type;

  @override
  String getDisplayValue(FieldNode dataObject) => '${getValue(dataObject)}';

  @override
  bool get supportsSorting => true;

  @override
  int compare(FieldNode a, FieldNode b) {
    final Comparable valueA = getValue(a);
    final Comparable valueB = getValue(b);
    return valueA.compareTo(valueB);
  }

  @override
  double get fixedWidthPx => 250.0;
}

class _FieldNameColumn extends ColumnData<FieldNode> {
  _FieldNameColumn() : super('Name');

  @override
  dynamic getValue(FieldNode dataObject) =>
      dataObject.isEmptyReference ? '' : dataObject.name;

  @override
  String getDisplayValue(FieldNode dataObject) => '${getValue(dataObject)}';

  @override
  bool get supportsSorting => true;

  @override
  int compare(FieldNode a, FieldNode b) {
    final Comparable valueA = getValue(a);
    final Comparable valueB = getValue(b);
    return valueA.compareTo(valueB);
  }

  @override
  double get fixedWidthPx => 150.0;
}

class _FieldValueColumn extends ColumnData<FieldNode> {
  _FieldValueColumn() : super('Value');

  @override
  dynamic getValue(FieldNode dataObject) =>
      dataObject.isEmptyReference || dataObject.isSentinelReference
          ? ''
          : dataObject.value;

  @override
  String getDisplayValue(FieldNode dataObject) {
    // TODO(bkonyi): is this necessary?
    /*if (dataObject is ObjectFieldNode && !dataObject.isNull) {
      // Real object that isn't Null value is empty string.
      return '';
    }*/

    // TODO(terry): Can we use standard Flutter string truncation?
    //              Long variable names the beginning and end of the name is
    //              most significant the middle part of name less so.  However,
    //              we need a nice hover/tooltip to show the entire name.
    //              If middle, '...' then we need a more robust computation w/
    //              text measurement so we accurately truncate the correct # of
    //              characters instead guessing that we only show 30 chars.
    var value = getValue(dataObject);
    if (value is String && value.length > 30) {
      value = '${value.substring(0, 13)}…${value.substring(value.length - 17)}';
    }
    return '$value';
  }

  @override
  bool get supportsSorting => true;

  @override
  int compare(FieldNode a, FieldNode b) {
    final Comparable valueA = getValue(a);
    final Comparable valueB = getValue(b);
    return valueA.compareTo(valueB);
  }

  @override
  double get fixedWidthPx => 250.0;
}
