/*// Copyright 2020 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import 'dart:async';

import 'package:flutter/material.dart';
import 'package:provider/provider.dart';
import 'package:vm_service/vm_service.dart';

import '../debugger/codeview.dart';
import '../debugger/debugger_controller.dart';
import '../debugger/debugger_model.dart';
import '../globals.dart';
import '../split.dart';
import '../table.dart';
import '../table_data.dart';
import '../utils.dart';
import '../vm_service_utils.dart';
import '../vm_service_wrapper.dart';
import 'vm_developer_common_widgets.dart';
import 'vm_developer_tools_screen.dart';
import 'vm_service_private_extensions.dart';
import 'vm_statistics_view_controller.dart';

class TestView extends VMDeveloperView {
  const TestView()
      : super(
          id,
          title: 'Test',
          icon: Icons.devices_sharp,
        );
  static const id = 'test';

  @override
  Widget build(BuildContext context) => TestViewBody();
}

/// Displays general information about the state of the Dart VM.
class TestViewBody extends StatelessWidget {
  final controller = VMStatisticsViewController();
  @override
  Widget build(BuildContext context) {
    return Column(
      children: [
        const Expanded(
          child: DartObjectInspector(object: null),
        ),
      ],
    );
  }
}

Future<Obj> getSomething() async {
  print('here');
  final service = serviceManager.service;
  final isolate = serviceManager.isolateManager.isolates.first;
  final classList = await service.getClassList(isolate.id);
  return await service.getObject(isolate.id,
      classList.classes.firstWhere((c) => c?.name == '_RawReceivePortImpl').id);
}

class DartObjectInspector extends StatelessWidget {
  const DartObjectInspector({@required this.object});

  final Obj object;

  @override
  Widget build(BuildContext context) {
    return FutureBuilder<Obj>(
      future: getSomething(),
      builder: (context, snapshot) {
        switch (snapshot.data?.type) {
          case 'Class':
            return _ClassInspector(
              clazz: snapshot.data as Class,
            );
          default:
            return const VMInfoCard(
              title: 'Instance of Closure',
              rowKeyValues: [
                MapEntry('Class', '_Closure'),
              ],
            );
        }
      },
    );
  }
}

class _ClassInspector extends StatelessWidget {
  _ClassInspector({@required this.clazz});

  final Class clazz;
  final _fieldTypeExpando = Expando<Instance>();

  Future<List<FieldRef>> _getFieldTypes(List<FieldRef> fields) async {
    final isolateId = serviceManager.isolateManager.selectedIsolate.id;
    await Future.wait(fields.map((e) => serviceManager.service
        .getObject(isolateId, e.declaredType.id)
        .then((type) => _fieldTypeExpando[e] = type)));
    return fields;
  }

  Future<_InstancesStatistics> _getInstancesStatistics() async {
    final service = serviceManager.service;
    final isolateId = serviceManager.isolateManager.selectedIsolate.id;

    final reachableSize = await service.getReachableSize(isolateId, clazz.id);
    final retainedSize = await service.getRetainedSize(isolateId, clazz.id);

    return _InstancesStatistics(
      reachableSize: int.parse(reachableSize.valueAsString),
      retainedSize: int.parse(retainedSize.valueAsString),
    );
  }

  @override
  Widget build(BuildContext context) {
    final isolateId = serviceManager.isolateManager.selectedIsolate.id;
    return FutureBuilder<Obj>(
      future: serviceManager.service.getObject(
        isolateId,
        clazz.classRef.id,
      ),
      builder: (context, snapshot) {
        final type = snapshot.data as Class;
        return FutureBuilder<Script>(
          future: serviceManager.service
              .getObject(isolateId, clazz.location.script.id)
              .then((e) => e as Script),
          builder: (context, snapshot) {
            final script = snapshot.data;
            return Expanded(
              child: Split(
                axis: Axis.horizontal,
                initialFractions: const [0.5, 0.5],
                children: [
                  Flexible(
                    child: Row(
                      children: [
                        Flexible(
                          child: Column(
                            children: [
                              VMInfoCard(
                                title: 'class ${clazz.name}',
                                rowKeyValues: _buildClassDetails(type, script),
                              ),
                              FutureBuilder<_InstancesStatistics>(
                                  future: _getInstancesStatistics(),
                                  builder: (context, snapshot) {
                                    final instancesStats = snapshot.data;
                                    print(instancesStats.retainedSize);
                                    return VMInfoCard(
                                      title: 'Instances',
                                      rowKeyValues: [
                                        const MapEntry(
                                            'Currently allocated', '0 (TODO)'),
                                        //MapEntry('Strongly ')
                                        MapEntry(
                                          'Reachable size',
                                          prettyPrintBytes(
                                              instancesStats.reachableSize,
                                              kbFractionDigits: 3,
                                              includeUnit: true),
                                        ),
                                        MapEntry(
                                          'Retained Size',
                                          prettyPrintBytes(
                                              instancesStats.retainedSize,
                                              kbFractionDigits: 3,
                                              includeUnit: true),
                                        ),
                                      ],
                                    );
                                  }),
                            ],
                          ),
                        ),
                        Flexible(
                          child: Column(
                            children: [
                              Flexible(
                                child: Card(
                                  child: SingleColumnTable<FuncRef>(
                                    title:
                                        'Functions (${clazz.functions.length})',
                                    data: clazz.functions,
                                    getNameCallback: (FuncRef f) => f.name,
                                    getValueCallback: (FuncRef f) => f.name,
                                    onItemSelected: (FuncRef f) {
                                      serviceManager.service
                                          .getObject(isolateId, f.id)
                                          .then(
                                        (value) {
                                          final func = value as Func;
                                          final controller =
                                              Provider.of<DebuggerController>(
                                                  context,
                                                  listen: false);
                                          controller.showScriptLocation(
                                            ScriptLocation(
                                              script,
                                              location: SourcePosition
                                                  .calculatePosition(
                                                script,
                                                func.location.tokenPos,
                                              ),
                                            ),
                                          );
                                        },
                                      );
                                    },
                                  ),
                                ),
                              ),
                              FutureBuilder<List<FieldRef>>(
                                future: _getFieldTypes(clazz.fields),
                                initialData: const [],
                                builder: (context, snapshot) {
                                  return Flexible(
                                    child: Card(
                                      child: SingleColumnTable<FieldRef>(
                                        title:
                                            'Fields (${snapshot.data.length})',
                                        data: snapshot.data,
                                        getNameCallback: (FieldRef f) => f.name,
                                        getValueCallback: (FieldRef f) {
                                          final nameBuffer = StringBuffer();
                                          if (f.isStatic) {
                                            nameBuffer.write('static ');
                                          }
                                          if (f.isConst) {
                                            nameBuffer.write('const ');
                                          }
                                          if (f.isFinal) {
                                            nameBuffer.write('final ');
                                          }
                                          nameBuffer.write(
                                              '${_fieldTypeExpando[f].name} ');
                                          nameBuffer.write(f.name);
                                          return nameBuffer.toString();
                                        },
                                        onItemSelected: (FieldRef f) async {
                                          final Field field =
                                              await serviceManager.service
                                                  .getObject(isolateId, f.id);
                                          final controller =
                                              Provider.of<DebuggerController>(
                                                  context,
                                                  listen: false);
                                          controller.showScriptLocation(
                                            ScriptLocation(
                                              script,
                                              location: SourcePosition
                                                  .calculatePosition(
                                                script,
                                                field.location.tokenPos,
                                              ),
                                            ),
                                          );
                                        },
                                      ),
                                    ),
                                  );
                                },
                              ),
                            ],
                          ),
                        )
                      ],
                    ),
                  ),
                  Column(
                    children: [
                      if (script != null)
                        Expanded(
                          child: CodeView(
                            controller:
                                Provider.of<DebuggerController>(context),
                            showHistory: false,
                            centerScrollingPosition: false,
                            initialPosition: ScriptLocation(
                              script,
                              location: SourcePosition.calculatePosition(
                                script,
                                clazz.location.tokenPos,
                              ),
                            ),
                            scriptRef: script,
                          ),
                        ),
                    ],
                  ),
                ],
              ),
            );
          },
        );
      },
    );
  }

  List<MapEntry> _buildClassDetails(Class type, Script script) {
    final sourceInfo =
        SourcePosition.calculatePosition(script, clazz.location.tokenPos);
    return [
      if (clazz.vmName != null && clazz.name != clazz.vmName)
        MapEntry('Internal Name', clazz.vmName),
      MapEntry('Class', ClassReference(type)),
      MapEntry('Library', LibraryReference(clazz?.library)),
      MapEntry('Script', ScriptReference(script, clazz)),
      if (clazz?.superClass != null)
        MapEntry('Superclass', ClassReference(clazz?.superClass)),
      if (clazz?.superType != null)
        MapEntry('Supertype', TypeReference(clazz.superType)),
      if (clazz?.mixin != null) MapEntry('Mixin', MixinReference(clazz.mixin)),
      if (clazz?.interfaces != null && clazz.interfaces.isNotEmpty)
        MapEntry('Implements',
            clazz.interfaces.map((i) => ClassReference(i.typeClass))),
      MapEntry(
          'Shallow Size',
          prettyPrintBytes(
            clazz.size,
            kbFractionDigits: 3,
            includeUnit: true,
          )),
    ];
  }
}

class _InstancesStatistics {
  const _InstancesStatistics(
      {@required this.reachableSize, @required this.retainedSize});

  final int reachableSize;
  final int retainedSize;
}

class _NameColumn<T> extends ColumnData<T> {
  _NameColumn({
    @required String title,
    @required this.getValueCallback,
  }) : super.wide(title);

  final String Function(T) getValueCallback;

  @override
  String getValue(T e) => getValueCallback(e);
}

class SingleColumnTable<T> extends StatelessWidget {
  SingleColumnTable({
    @required this.title,
    @required this.data,
    @required this.getNameCallback,
    @required String Function(T) getValueCallback,
    this.onItemSelected,
  }) : name = _NameColumn<T>(
          title: title,
          getValueCallback: getValueCallback,
        );

  final String title;
  final List<T> data;
  final String Function(T) getNameCallback;
  final void Function(T) onItemSelected;

  final _NameColumn<T> name;

  List<ColumnData<T>> get columns => [
        name,
      ];

  @override
  Widget build(BuildContext context) {
    return FlatTable<T>(
      columns: columns,
      data: data,
      keyFactory: (T e) => ValueKey<String>(getNameCallback(e)),
      sortColumn: name,
      sortDirection: SortDirection.ascending,
      onItemSelected: onItemSelected,
    );
  }
}
*/